"use strict";
// ♻️
// IMPORTANT: Take care to apply changes to all supported languages
// when modifying code or documentation in the services runtime.
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const ServiceChannel_1 = require("./ServiceChannel");
const Helpers_1 = require("./Helpers");
const Debugging_1 = require("./Debugging");
/**
 * Manages a cluster of channels and services and sets up service communication. Use the auto-generated Service objects
 * to register implementations and get proxies to cross-process services. Most clients will set up a single shared object
 * with multiple channels.
 */
class ServiceManager {
    constructor() {
        /**
         * Creates a new proxy for a given service. Since the presence of the service isn't checked, errors may occur when
         * actually calling methods on the proxy.
         *
         * @param service The auto-generated definition object for the wanted service
         * @param channel Where the service is assumed to live. Calling this method implicitly registers the channel.
         */
        this.expect = (service, channel) => {
            debug.log("ServiceManager.expect:", service.id);
            return this.addRouter(channel).expect(service);
        };
        /** Removes a registered implementation or channel. */
        this.unregister = (_) => {
            // FIXME: implement
            throw new Error("unregister isn't implemented");
        };
        this.routers = new Set();
        this.addRouter = (channel) => {
            for (const existing of this.routers) {
                if (existing.channel === channel) {
                    return existing;
                }
            }
            const router = new Router(channel);
            this.routers.add(router);
            return router;
        };
    }
    register(what) {
        return __awaiter(this, void 0, void 0, function* () {
            function isImplementation(obj) {
                return obj.service !== undefined && obj.implementation !== undefined;
            }
            if (isImplementation(what)) {
                debug.log("ServiceManager.registerImplementation:", what.service.id);
                yield this.addRouter(what.channel).registerImplementation(what.implementation, what.service);
            }
            else {
                debug.log("ServiceManager.registerChannel");
                this.addRouter(what.channel);
            }
        });
    }
}
exports.ServiceManager = ServiceManager;
(function (ServiceManager) {
    /**
     * Shared service manager for TypeScript clients that only have one area of responsibility within the runtime,
     * exposed as the generated `SomeService.on(channel)` convenience API. If your current script runtime only has
     * a single domain of responsibility (e.g. Vekter representing a single document), use the convenience instead
     * of constructing your own service managers.
     */
    function shared() {
        if (ServiceManager.isTesting) {
            if (!testingServiceManager) {
                throw new Error("ServiceManager.shared() may not be used while testing. Use testWithShared() for explicitness.");
            }
            return testingServiceManager;
        }
        return sharedServiceManager;
    }
    ServiceManager.shared = shared;
    /** Enables or disables testing mode. */
    ServiceManager.isTesting = false;
    /** Temporarily changes the value of ServiceManager.shared(), for testing purposes. */
    function testWithShared(manager, fn) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!ServiceManager.isTesting)
                throw new Error("ServiceManager.isTesting must be true to use testWithShared()");
            if (testingServiceManager)
                throw new Error("ServiceManager.testWithShared() may not be nested");
            testingServiceManager = manager;
            yield fn();
            testingServiceManager = undefined;
        });
    }
    ServiceManager.testWithShared = testWithShared;
    const sharedServiceManager = new ServiceManager();
    let testingServiceManager;
})(ServiceManager = exports.ServiceManager || (exports.ServiceManager = {}));
class Router {
    constructor(channel) {
        this.channel = channel;
        this.onMessage = (message) => {
            debug.log("Router.onMessage:", message);
            try {
                // For now, just assume we're in a wonderful world where a type check implies a complete and valid message
                if (message.type === ServiceChannel_1.ServiceChannel.MessageType.Request) {
                    void this.onRequest(message);
                }
                else if (message.type === ServiceChannel_1.ServiceChannel.MessageType.Response) {
                    if (message.method === Discovery.method) {
                        this.onDiscoveryResponse(message);
                    }
                    else {
                        this.onResponse(message);
                    }
                }
                else {
                    throw new Error(`Unknown service router message: ${JSON.stringify(message)}`);
                }
            }
            catch (error) {
                debug.log("Router.onMessage error:", error);
            }
        };
        // Discovery
        this.onDiscoveryResponse = (response) => {
            // TODO: currently, TypeScript only sends discovery info (to Swift) but doesn't accept it
        };
        this.broadcastDiscoveryInfo = () => {
            const services = {};
            for (const key in this.implementedServices) {
                const service = this.implementedServices[key].service;
                services[key] = { fingerprint: service.fingerprint };
            }
            const discoveryInfo = { services };
            try {
                this.channel.postMessage({
                    type: ServiceChannel_1.ServiceChannel.MessageType.Response,
                    id: "",
                    serviceId: "",
                    method: Discovery.method,
                    body: discoveryInfo,
                });
            }
            catch (_a) {
                // postMessage can fail (e.g. when generating a "dehydrated" version of the host page). Since this broadcast happens automatically and not as part of explicitly calling a service method, suppress errors here. If discovery fails, the actual services will fail as well.
            }
        };
        // Outgoing messages
        this.expect = (service) => {
            return service.newOutgoingWrapper((message) => __awaiter(this, void 0, void 0, function* () {
                return new Promise(resolve => {
                    debug.log("Router.send:", message);
                    const requestId = Helpers_1.ServiceRuntimeHelpers.generateUniqueId();
                    this.waitingRequestsMap[requestId] = resolve;
                    this.channel.postMessage({
                        type: ServiceChannel_1.ServiceChannel.MessageType.Request,
                        id: requestId,
                        serviceId: service.id,
                        method: message.method,
                        stream: StreamOperation.toMessage(message.stream),
                        body: message.argument,
                    });
                });
            }));
        };
        this.waitingRequestsMap = {};
        this.onResponse = (response) => {
            const resolve = this.waitingRequestsMap[response.id];
            if (!resolve)
                return debug.error("Router received response without corresponding request:", response);
            delete this.waitingRequestsMap[response.id];
            resolve(response.body);
        };
        // Incoming messages
        this.registerImplementation = (rawImplementation, service) => __awaiter(this, void 0, void 0, function* () {
            debug.log("Router.registerImplementation:", rawImplementation);
            // Only expose known methods, bound to the implementation object so they can be called as is
            const implementation = {};
            for (const key in service.methods) {
                const name = key;
                const method = rawImplementation[name];
                if (typeof method !== "function") {
                    throw new Error(`Implementation for ${service.id} doesn't correctly implement ${name}()`);
                }
                implementation[name] = method.bind(rawImplementation);
            }
            this.implementedServices[service.id] = { service, implementation };
            this.broadcastDiscoveryInfo();
        });
        this.implementedServices = {};
        this.streamIterators = {};
        this.onRequest = (request) => __awaiter(this, void 0, void 0, function* () {
            debug.log("Router.onRequest:", request);
            const implementedService = this.implementedServices[request.serviceId];
            const implementation = implementedService ? implementedService.implementation : undefined;
            // FIXME: return error response
            if (!implementation)
                return debug.error("Router received unknown request:", request);
            try {
                let resultBody;
                if (request.stream) {
                    const stream = StreamOperation.fromMessage(request.stream);
                    let iterator = this.streamIterators[stream.id];
                    if (!iterator) {
                        // Start a new streaming iterator if one doesn't exist yet
                        const method = implementation[request.method];
                        const iteratorStream = (yield method(request.body));
                        iterator = iteratorStream[Symbol.asyncIterator]();
                        this.streamIterators[stream.id] = iterator;
                    }
                    if (stream.method !== "next") {
                        throw new Error("Stream operations other than next() are not yet supported");
                    }
                    const iteratorResult = yield iterator.next();
                    resultBody = { done: iteratorResult.done, value: iteratorResult.value };
                }
                else {
                    const method = implementation[request.method];
                    resultBody = yield method(request.body);
                }
                // Note: the result type isn't verified or validated in any way
                this.channel.postMessage({
                    type: ServiceChannel_1.ServiceChannel.MessageType.Response,
                    id: request.id,
                    serviceId: request.serviceId,
                    method: request.method,
                    body: resultBody,
                });
            }
            catch (error) {
                // FIXME: send a throw response
                debug.log("Router.onRequest error:", error.message);
            }
        });
        channel.addMessageListener(this.onMessage);
    }
}
var Discovery;
(function (Discovery) {
    Discovery.method = "#discover";
})(Discovery || (Discovery = {}));
var StreamOperation;
(function (StreamOperation) {
    const returnPrefix = "#return:";
    const throwPrefix = "#throw:";
    function fromMessage(operation) {
        if (operation.startsWith(returnPrefix)) {
            return { id: operation.substr(returnPrefix.length), method: "return" };
        }
        else if (operation.startsWith(throwPrefix)) {
            return { id: operation.substr(throwPrefix.length), method: "throw" };
        }
        else {
            return { id: operation, method: "next" };
        }
    }
    StreamOperation.fromMessage = fromMessage;
    function toMessage(stream) {
        if (!stream)
            return undefined;
        switch (stream.method) {
            case "next":
                return stream.id; // Having just a stream identifier implicitly means calling next()
            case "return":
                return returnPrefix + stream.id;
            case "throw":
                return throwPrefix + stream.id;
            default:
                return undefined;
        }
    }
    StreamOperation.toMessage = toMessage;
})(StreamOperation || (StreamOperation = {}));
const debug = Debugging_1.ServiceDebugging.createLogger("📮");
